import logging
from math import ceil
from functools import lru_cache
from typing import Iterable
from multiprocessing.pool import ThreadPool
from swarkn.helpers import swallow_exception_wrapper

logger = logging.getLogger(__name__)

@lru_cache()
def _get_pool():
    return ThreadPool(10)


def chunk(iterable: Iterable, size: int):
    return (iterable[i:i + size] for i in range(0, len(iterable), size))

def prl_map(func: callable, iterable: Iterable, chunksize=None, pool=None, ignore_failures=False):
    pool = pool or _get_pool()
    chunksize = chunksize or ceil(len(iterable) / len(pool._pool))
    chunks = list(chunk(iterable, chunksize))
    logger.info(f'splitting {len(iterable)} elements into {len(chunks)} buckets of size {chunksize}')
    wrapped_func = swallow_exception_wrapper(func) if ignore_failures else func
    res = pool.map(wrapped_func, chunks)
    return res

def freeze(x):
    from frozendict import frozendict
    return frozendict(x) if isinstance(x, dict) else x

def unfreeze(x):
    return dict(x) if isinstance(x, Mapping) else x


import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import re
from itertools import cycle

DF = pd.DataFrame

def stack(pivoted: DF) -> DF:
    df = pivoted.stack().reset_index(1)
    df.index = pd.to_datetime(df.index)
    df.columns = ['variable', 'value']
    return df


def cfg2subgraph(subplots: dict[list[dict]],
                 dfs: DF | dict[str, DF],
                 chart_type='Scatter',
                 var_col='variable',
                 val_col='value',
                 text_col: str = None,
                 customdata_cols: list = None,
                 override_func=lambda *_, **__: {},
                 height=600,
) -> go.Figure:
    ### setup ###
    dfs = dfs if isinstance(dfs, dict) else {'': dfs}
    keys = tuple(f'{dfkey} {cfgkey}' for cfgkey in subplots for dfkey in dfs)

    #### create subplot #####
    fig = make_subplots(len(keys),
        shared_xaxes=True,
        vertical_spacing=0.05,
        # row_heights=[CHART_SIZE] * len(subplots),
        subplot_titles=keys,
    )

    ### add traces ####
    subplot_row = 1
    for subplot in subplots.values():
        for dfkey, df in dfs.items():
            for cfg in subplot:
                chart_type = cfg.get('type', chart_type)
                kwargs_ = cfg.get('kwargs', {})
                kwargs_func = lambda *x: kwargs_ if isinstance(kwargs_, dict) else kwargs_(*x)
                for col, subdf in df.query(f"{var_col}.str.contains(@cfg['regex'])").groupby(var_col):
                    kwrgs = {
                        'hoverlabel': {'namelength': -1},
                        'name': col,
                        'customdata': subdf[customdata_cols] if customdata_cols else None,
                        'text': subdf[text_col] if text_col else None,
                        'meta': dict(col=col, dfkey=dfkey),
                    } | kwargs_func(col, dfkey) | override_func(col, dfkey)

                    gob = getattr(go, chart_type)(x=subdf.index, y=subdf[val_col], **kwrgs)
                    fig.add_trace(gob, row=subplot_row, col=1)
            subplot_row += 1

    # TODO: reorder legend into original input df order
    fig.update_xaxes(showticklabels=True)  # , tickangle=-45
    fig.update_layout(height=subplot_row * height, legend=dict(
        groupclick="toggleitem")
    )

    return fig

################################# CHART COLOURS #########################################



if __name__ == '__main__':
    from swarkn.charting.colors import color_cfg, cycle_colors
    from functools import partial
    import plotly.io as pio
    import pandas as pd
    pio.renderers.default = "browser"

    cfg = {
        'curves1': [
            {
                'regex': 'curves1_',
                'colors': partial(color_cfg, {}),
                'kwargs': {'line': {"dash": "dot"}},
            }, {
                'regex': 'curves1a_',
                #'type': 'Bar',
            },
        ],
        'curves2': [
            {
                'regex': 'curves2',
                'colors': cycle_colors,
                'kwargs': lambda col, dfkey, *_: dict(
                    line={"dash": "dot"},
                    hovertemplate='%{text} %{customdata}',
                    legendgroup='curves2',
                    legendgrouptitle_text='curve2_title',
                ),
            },
        ],
    }

    df = pd.DataFrame({
        'category': ['a'] * 5 + ['b'] * 5,
        'category_text': ['aa'] * 4 + ['bb'] * 6,
        "curves1_xx": list(range(10)),
        "curves1a_xx": list(range(10, 20)),
        "curves2_xx": list(range(10, 20)),
    }).reset_index().melt(['index', 'category', 'category_text']).set_index('index')
    fig = cfg2subgraph(cfg, df,
        # text_col='category_text',
        # customdata_cols='category',
    )

    fig.show()


import re

def dict2predicate_sql(predicates: dict, andor='AND') -> str:
    res = []
    for k, v in predicates.items():
        value = v
        if isinstance(v, (tuple, set, list)):
            operator = 'in'
            value = re.sub('\]|\}', ')',
                re.sub('\[|\{', '(', str(v))
            )
        elif isinstance(v, str):
            operator = 'like'
            value = f"'{v}'"
        elif v is None:
            operator = 'is'
            value = 'Null'
        else:
            operator = '='

        if k.startswith('~'):
            operator = '!=' if operator == '=' else \
                       'is not' if operator == 'is' else \
                       f'not {operator}'

        res.append(f"{k.replace('~', '')} {operator} {value}")
    return f' {andor} '.join(res)




import boto3
import pyarrow as pa
import pyarrow.dataset as ds

def session_storage_options(session: boto3.Session = None) -> dict:
    """
    convenience function used in DeltaTable
    """
    if not session:
        boto3.setup_default_session()
        session = boto3.DEFAULT_SESSION
    creds = session.get_credentials()

    storage_options = dict(
        AWS_REGION=session.region_name,
        AWS_ACCESS_KEY_ID=creds.access_key,
        AWS_SECRET_ACCESS_KEY=creds.secret_key,
        AWS_S3_ALLOW_UNSAFE_RENAME='true'
    )
    return storage_options

def parquet_dataset(path: str) -> ds.Dataset:
    return ds.dataset(path.replace('s3://', ''),
        format="parquet",
        filesystem=pa.fs.S3FileSystem(),
        partitioning="hive"
    )


import re
from itertools import cycle


def color_cfg(cfg: dict, col: str):
    for regex, color in cfg.items():
        if re.search(regex, col):
            return color
    return None

def cycle_colors(colors: list):
    cyc = cycle(colors)
    return lambda *_: next(cyc)




from swarkn.imports import load_module
from swarkn.helpers import run_func
def test_load_module():
    path = 'swarkn.db.sql'
    path = r'C:\Users\user\PycharmProjects\pysak\swarkn\db\sql.py'
    module = load_module(path)
    res = module.dict2predicate_sql()

    run_func(module.dict2predicate_sql, predicates={'a': 1}, andor='OR')



import logging
import json
import pandas as pd
from contextlib import contextmanager
from functools import partial
from typing import Iterable, Union, List, Dict
from psycopg2.extras import execute_values, Json
from psycopg2.pool import ThreadedConnectionPool
from psycopg2.sql import SQL, Identifier, Composed, Literal, Composable

from pythonlib.utils.helpers import get_env, json_serial, wraplist
from pythonlib.utils.pandas_helpers import filter_dict

SQL_ITERABLES = (list, tuple, set, frozenset)


class ThreadedConnectionPoolAug(ThreadedConnectionPool):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(__name__)

    @contextmanager
    def cursor(self):
        try:
            _conn = self.getconn()
            with _conn as real_conn:
                cursor = real_conn.cursor()
                yield cursor
                self.logger.info(f'{cursor.rowcount} rows affected')
                real_conn.commit()
        except:
            raise
        finally:
            _conn.rollback()
            self.putconn(_conn)

def cols_from_iterable(cols: Iterable, template='{}'):
    return SQL(', ').join([SQL(template).format(Identifier(col)) for col in cols])

def predicates_from_dict(predicates: dict, operator=SQL(' AND ')) -> SQL:
    """

    tries to cleverly infer predicates from a dict
    handles LIKE, IN and =
    """
    res = Composed([SQL(' WHERE '),
                    operator.join([
                        SQL(' ').join([
                            Identifier(key),
                            SQL(sql_operator(value)),
                            Literal(tuple(value) if isinstance(value, SQL_ITERABLES) else value),
                        ]) for key, value in predicates.items()
                    ])
                    ]) if predicates else SQL('')
    return res


def sql_operator(value) -> str:
    """

    Try to cleverly infer the operator based on literal
    """
    res = 'LIKE' if isinstance(value, str) and '%' in value \
        else 'IS' if value is None \
        else 'IN' if isinstance(value, SQL_ITERABLES) \
        else '='
    return res


def _cast_inputs(list_of_tuples: list):
    res = [[
        Json(elem, dumps=partial(json.dumps, default=json_serial)) if isinstance(elem, dict) else elem
        for elem in tup
    ] for tup in list_of_tuples]
    return res


class DBHelper:
    MAX_POOLSIZE = 8
    MIN_POOLSIZE = 1

    def __init__(self, **login_params):
        self.pool = ThreadedConnectionPoolAug(self.MIN_POOLSIZE, self.MAX_POOLSIZE, **login_params)
        self.logger = logging.getLogger(__name__)

    def execute(self, sql: Composable):
        with self.pool.cursor() as cursor:
            self.logger.info(f'{sql.as_string(cursor)}')
            cursor.execute(sql)
            return cursor.rowcount


    def insert_json(self, table: str, values, schema='public', **kwargs):
        """

        :param values: dict or list of dicts
        :param kwargs: will be added to table
        """
        list_of_values = wraplist(values)
        jsonified_values = [(
            dict(value, **kwargs),  # this is single element tuple
        ) for value in list_of_values]
        latestid = self.insert(table, jsonified_values, schema, returning='id')
        return latestid

    def insert(self, table: str, values, schema='public', returning: str = None,
               cols: Union[list, tuple] = None, #uses table col order by default
               upsertcols: Iterable = None,
        ):
        """

        :param values: list of tuples. or single tuple.
        :param upsertcols: set of keys in "values"
        """
        list_of_values = _cast_inputs(wraplist(values))
        colsql = SQL("({})").format(cols_from_iterable(cols)) if cols else ''
        if upsertcols:
            nonpks = set(cols) - set(upsertcols)
            conflict = SQL("ON CONFLICT({}) DO UPDATE SET {};").format(
                cols_from_iterable(upsertcols),
                cols_from_iterable(nonpks, '{0} = EXCLUDED.{0}'),
            )
        else:
            conflict = None

        insert_str = Composed(filter(None, [
            SQL("insert into {}.{} ").format(Identifier(schema), Identifier(table)),
            colsql,
            SQL(" values %s "),
            conflict,
            SQL("RETURNING {}").format(Identifier(returning)) if returning else None
        ]))

        with self.pool.cursor() as cursor:
            self.logger.info(f'{insert_str.as_string(cursor)} ({len(list_of_values)} rows)')
            execute_values(cursor, insert_str, list_of_values)
            if returning:
                ids = cursor.fetchall()
                return ids[0][0] #latest id

    def update(self, table: str,
               col_values: dict,
               schema='public',
               predicates={}
               ):
        # col_values = {'1':  2, '3', 4}
        # table = 'test'
        update_str = Composed([
            SQL("UPDATE {}.{} SET ").format(Identifier(schema), Identifier(table)),
            SQL(", ").join(SQL('{}={}').format(Identifier(key), Literal(val)) for key, val in col_values.items()),
            predicates_from_dict(predicates),
        ])
        self.logger.info(f'updating {len(col_values)} cols')
        rowcount = self.execute(update_str)
        return rowcount

    def delete(self, table: str,
               schema='public',
               predicates={}
               ):
        sqlstr = Composed([
            SQL("delete from {}.{} ").format(Identifier(schema), Identifier(table)),
            predicates_from_dict(predicates),
        ])
        rowcount = self.execute(sqlstr)
        return rowcount


    def get_table(self, table: str,
                  schema='public',
                  limit=99999999,
                  predicates={},
                  cols=None,
                  session_vars={}) -> pd.DataFrame:
        colstr = SQL(',').join(map(Identifier, cols)) if cols else SQL('*')
        query = Composed([
            SQL("set {} = {}; ").format(Identifier(var), Literal(val))
        for var, val in session_vars.items()] +
        [
            SQL("select {} from {}.{} ").format(colstr, Identifier(schema), Identifier(table)),
            predicates_from_dict(predicates),
            SQL(" limit {}").format(Literal(limit))
        ])
        with self.pool.cursor() as cursor:
            self.logger.info(query.as_string(cursor))
            df = pd.read_sql(query, cursor.connection)
        return df

    def get_json_table(self, *args, predicates=(), cols=None, **kwargs) -> pd.DataFrame:
        """

        expands single columnn jsonb table into individual columns
         :param args: see self.get_table()
        """
        df = self.get_table(*args, **kwargs)
        dfj = pd.concat([
            df[['id']],
            pd.io.json.json_normalize(df['jblob'])
        ], axis=1)
        df_filtered = filter_dict(dfj, predicates)
        return df_filtered[cols] if cols else df_filtered

    def get_table_as_dict(self, *args, json_table=False, **kwargs) -> List[Dict]:
        """

        :param args: see self.get_table()
        """
        func = self.get_json_table if json_table else self.get_table
        df = func(*args, **kwargs)
        res = json.loads(df.to_json(orient='records')) #cleans up Nans and other non json constructs
        return res


if __name__ == '__main__':
    from datetime import date

    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')

    dbhelper = DBHelper()
    #    latestid = dbhelper.insert_json('json', [{'a': 1}], 'test')
    #   latestid = dbhelper.insert('json', [(Json({'dsfdsf': 1}), 206)], 'test')
    dbhelper.update('json', [('id', 999)], 'test', {'id': 1000})
    df = dbhelper.get_table('json', 'test')
    print(df)

    preds = {
        "hello": "world",
        "hello1": "world%",
        "hello2": 1,
        "hello3": [11, 3, 4],
        "hello4": ('dsfdsf',),
        "hello5": date(2018, 1, 1),
    }
    predstr = predicates_from_dict(preds)

    self = dbhelper
    with dbhelper.pool.cursor() as cursor:
        cursor.execute(
            '''
CREATE OR REPLACE VIEW public.v_users AS
    SELECT t.*
        , kyc_gbg and coalesce(kyc_pep, true) and coalesce(kyc_override, true) as kyc_pass
    from raw.users t
            '''
        )
        res = cursor.fetchall()
        res

        self.logger.info(query.as_string(conn))

    res = dbhelper.get_table('users', 'raw', predicates={'id': 4})
    res['jblob'][0]


import frozendict
import sys
import logging
from typing import Callable
from collections.abc import Mapping
from functools import lru_cache
from time import perf_counter
from contextlib import contextmanager
logger = logging.getLogger(__name__)

@contextmanager
def timer(msg=None) -> float:
    start = perf_counter()
    yield perf_counter() - start
    cost = perf_counter() - start
    msg = msg or "Time taken: {cost}s"
    logger.info(msg.format(cost=cost))

@lru_cache()
def is_local():
    return 'win' in sys.platform

def init_logger(level=logging.INFO, format='%(asctime)s|%(name)s|%(levelname)s|%(message)s'):
    logging.basicConfig(level=level, format=format)

def swallow_exception_wrap(func, return_exception=False, **logger_kwags, ):
    def innerfunc(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.error(f'failed func: {func} args: {args}, kwargs: {kwargs}, exception: {e}', **logger_kwags)
            if return_exception:
                return e

    return innerfunc

@contextmanager
def swallow_exception(**kwags):
    try:
        yield
    except Exception as ex:
        logger.error(ex, **kwags)


def freeze(x):
    from frozendict import frozendict
    return frozendict(x) if isinstance(x, dict) else x

def unfreeze(x):
    return dict(x) if isinstance(x, Mapping) else x

def clean_kwargs(fn: Callable, kwargs: dict) -> dict:
    params = kwargs if 'kwargs' in fn.__code__.co_varnames else {
        k: v for k, v in kwargs.items() if k in fn.__code__.co_varnames
    }
    return params

def run_func(fn: Callable, *args, **kwargs):
    """

    removes irrelevant kwargs before calling fn
    """
    params = clean_kwargs(fn, kwargs)
    return fn(*args, **params)




import sys
from importlib.util import spec_from_file_location, module_from_spec
from importlib import import_module


def load_module(path: str, module_name="module.name"):
    """

    path = '/home/test'
    path = 'c:/users/user/test.py'
    """
    if path.endswith('.py'):
        spec = spec_from_file_location(module_name, path)
        module = module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
        return module
    return import_module(path)